import argparse
import re
import datetime
import pynusmv as pn
from tools.ctl import eval as ctl

def _current_time():
    """
    Return the current time (YY-MM-DD HH:MM:SS).
    """
    return datetime.datetime.strftime(datetime.datetime.now(),
                                      "[%y-%m-%d %H:%M:%S]")

def verbose(*args, **kwargs):
    """
    Print the given arguments, adding the current time in front of them and
    flushing the printing.
    """
    print(_current_time(), *args, flush=True, **kwargs)

def nbroutes(fsm):
    """
    Return the number of routes of fsm, that is, the highest integer X
    such that there exists a variable RX.st in fsm.
    
    If RX.st exists, the model is assumed to have also RY.st for 1 <= Y <= X
    (that is, the highest integer X defines the number of routes of the model).
    """
    renumber = re.compile("R([0-9]+).st")
    nbroutes = 0
    for var in fsm.bddEnc.stateVars:
        match = renumber.search(var)
        if match is not None:
            route = int(match.group(1))
            if nbroutes < route:
                nbroutes = route
    return nbroutes

def extract_routes(fsm, nbroutes, states):
    """
    Return the set of subset of routes that can be set at the same time in
    states.
    
    fsm -- the model;
    nbroutes -- the number of routes of the model;
    states -- a BDD-encoded subset of states of fsm from which extract the
              different subsets of routes.
    
    Return a set of set of route names (e.g. R1) such that for any set S of
    this set, there exists a state st in states such that exactly the routes
    belonging S are setted in st (i.e. RX.st = s).
    """
    setted_routes = set()
    
    # The set of variables representing the state of the routes
    rstates = {'R%d.st' % i for i in range(1, nbroutes + 1)}
    
    rcube = fsm.bddEnc.cube_for_state_vars(rstates)
    ocube = fsm.bddEnc.statesCube - rcube
    
    while states.isnot_false():
        s = fsm.pick_one_state(states)
        
        # The set of routes enabled in s
        enabled = {var[:var.find(".")]
                   for var, val in s.get_str_values().items()
                   if var in rstates and val == "s"}
        setted_routes.add(frozenset(enabled))
        
        states = states - s.forsome(ocube)
    
    return setted_routes

def compatibility_table(setted_routes, nbroutes=None):
    """
    Given a set of sets of routes that can be setted at the same time,
    return the compatibility table, that is, for each pair of routes, whether
    or not they can be setted at the same time.
    
    setted_routes -- a set of sets of route names (e.g. R1);
    nbroutes -- the number of routes to consider;
                at least the maximum route number present in setted_routes.
    
    If nbroutes is None, the number of routes to consider is the maximum route
    number present in setted_routes.
    
    Return a list of lists, where the top list and the sub-lists have length
    nbroutes, and where the value of the j-1th element of the i-1th list is
    True if Ri can be setted at the same time as Rj, and False otherwise.
    """
    if nbroutes is None:
        nbroutes = 0
        for subset in setted_routes:
            for r in subset:
                nbroutes = max(nbroutes, int(r[1:]))
    table = [[False] * nbroutes for _ in range(nbroutes)]
    for subset in setted_routes:
        for r in subset:
            ri = int(r[1:])
            for rr in subset:
                rri = int(rr[1:])
                table[ri-1][rri-1] = True
                table[rri-1][ri-1] = True
    return table

def print_compatibility_table(table):
    nbroutes = len(table)
    print(" " * 5 + 
          "".join("{:>5}".format("R" + str(i))
                  for i in range(1, nbroutes + 1)))
    for i in range(nbroutes):
        print("{:<5}".format("R" + str(i + 1)) +
              "".join("{:>5}".format("V" if table[i][j] else "X")
                      for j in range(nbroutes)))

def compute_deadlock_states(fsm, states=None):
    """
    Return the BDD representing the subset deadlock states in fsm. If states is
    not None, return the subset of states that cannot lead to any state of
    states in one step.
    
    fsm -- the model;
    states -- if not None, a BDD-encoded subset of fsm states.
    """
    if states is None:
        # If states is None, we need to check reachable states
        # fsm.pre(true()) should be easier to compute than fsm.pre(reachable)
        states = pn.dd.BDD.true(fsm)
        no_succ = ~fsm.pre(states) & fsm.state_constraints
        return fsm.reachable_states & no_succ & fsm.bddEnc.statesMask
    else:
        no_succ = ~fsm.pre(states) & fsm.state_constraints
        return states & no_succ & fsm.bddEnc.statesMask

def check_deadlock_free(fsm, states=None):
    """
    Return whether fsm is deadlock-free. If states is not None, the check is
    made on the sub-system defined by state.
    
    fsm -- the model;
    states -- if not None, a BDD-encoded subset of fsm states.
    """
    return compute_deadlock_states(fsm, states).is_false()

def check_on_subset(fsm, states, spec, ignore_fairness=False):
    """
    Return the subset of states that satisfy spec in the sub-part of fsm
    defined by states. states should define a sub-system with no deadlock.
    
    fsm -- the model;
    states -- a BDD-encoded set of states of fsm;
    spec -- a CTL specification.
    ignore_fairness -- whether or not using CTL algorithms ignoring fairness
                       constraints.
    """
    oldreach = fsm.reachable_states
    fsm.reachable_states = states
    result = check_ctl(fsm, spec, ignore_fairness=ignore_fairness)
    fsm.reachable_states = oldreach
    return result

def compute_not_ag_ef_init(fsm, ignore_fairness=False, cache={}):
    """
    Return the subset of reachable states of fsm not satisfying AG EF INIT
    where INIT represents fsm initial states.
    
    fsm -- the model;
    ignore_fairness -- whether or not using CTL algorithms ignoring fairness
                       constraints;
    cache -- the cache of the function, should stay default argument.
    
    The returned value is the BDD representing the set of reachable
    states from which init cannot be reached.
    """
    if fsm not in cache:
        if ignore_fairness:
            efi = ctl.eu(fsm, pn.dd.BDD(fsm), fsm.init)
        else:
            efi = pn.mc.ef(fsm, fsm.init)
        unsat_states = fsm.reachable_states & fsm.state_constraints & ~efi
        cache[fsm] = unsat_states & fsm.bddEnc.statesMask
    return cache[fsm]

def check_invariant(fsm, spec):
    """
    Return whether spec is an invariant of fsm reachable states or not.
    
    fsm -- the model;
    spec -- a str-encoded simple expression on variables of the model
            or a BDD.
    """
    if isinstance(spec, str):
        spec = pn.mc.eval_simple_expression(fsm, spec)
    return fsm.reachable_states & fsm.state_constraints <= spec

def check_reachability(fsm, spec):
    """
    Return whether spec is reachable from the initial states of fsm,
    that is, whether there exists a reachable state in fsm satisfying spec.
    
    fsm -- the model;
    spec -- a str-encoded simple expression on variables of the model
            or a BDD.
    """
    if isinstance(spec, str):
        spec = pn.mc.eval_simple_expression(fsm, spec)
    return (fsm.reachable_states & fsm.state_constraints & spec).isnot_false()

def check_ctl(fsm, spec, agefi=False, ignore_fairness=False):
    """
    Return whether spec is satisfied by the initial states of fsm or not.
    
    fsm -- the model;
    spec -- a CTL specification;
    agefi -- if True, the model is assumed to satisfy AG EF INIT;
    ignore_fairness -- whether or not using CTL algorithms ignoring fairness
                       constraints.
    
    For optimisation reasons, the spec must be out of context (main context)
    and top context will be ignored.
    """
    # Ignore top context
    if spec.type == pn.nusmv.parser.parser.CONTEXT:
        spec = spec.cdr
    
    if spec.type == pn.nusmv.parser.parser.AG:
        spec = spec.car
        
        if spec.type == pn.nusmv.parser.parser.EF and agefi:
            # Optimisation:
            # AG EF p is equivalent to EF p
            # if the model satisfies AG EF init
            # EF p is equivalent to checking reachability of p
            return check_reachability(fsm, str(spec.car))
        
        else:
            # Optimisation
            # AG p is equivalent to checking invariance of p
            if ignore_fairness:
                sat = ctl.eval_ctl(fsm, spec)
            else:
                sat = pn.mc.eval_ctl_spec(fsm, spec)
            return check_invariant(fsm, sat)
    else:
        if ignore_fairness:
            sat = ctl.eval_ctl(fsm, spec)
        else:
            sat = pn.mc.eval_ctl_spec(fsm, spec)
        return (fsm.init & fsm.state_constraints & ~sat).is_false()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Interlocking tester.",
                                     epilog="If no optional argument is given "
                                            "(appart from -h, -v, -dyn), "
                                            "all checks are performed.")
    # Populate arguments
    parser.add_argument("model", help="the SMV model to check")
    parser.add_argument("-v", action="store_true", dest="verbose",
                        help="activate verbose mode")
    parser.add_argument("-dyn", action="store_true", dest="dynamic",
                        help="activate dynamic BDD reordering")
    parser.add_argument("-ord", dest="ordering", default=None,
                        help="specify an initial ordering for BDD variables")
    parser.add_argument("-i", action="store_true", dest="init",
                        help="count and print all initial states")
    parser.add_argument("-r", action="store_true", dest="reach",
                        help="explicitely compute reachable states "
                             "and count them")
    parser.add_argument("-m", action="store_true", dest="machine",
                        help="check the fsm consistency")
    parser.add_argument("-d", action="store_true", dest="deadlock",
                        help="check the model for deadlocks")
    parser.add_argument("-c", action="store_true", dest="compatibility",
                        help="compute the routes compatibility table "
                             "and print it")
    parser.add_argument("-agefi", action="store_true", dest="agefi",
                        help="check AG EF init on the model")
    parser.add_argument("-sagefi", action="store_true", dest="sagefi",
                        help="check AG EF init on the model and print the set "
                             "of reachable states not satisfying it")
    parser.add_argument("-fr", action="store_true", dest="fair_reach",
                        help="check that all reachable states are fair states")
    parser.add_argument("-if", action="store_true", dest="ignore_fairness",
                        help="use CTL algorithms ignoring fairness "
                             "constraints")
    parser.add_argument("-inv", action="store_true", dest="invariants",
                        help="check the invariants of the model")
    parser.add_argument("-ctl", action="store_true", dest="ctl",
                        help="check CTL specifications of the model")
    args = parser.parse_args()
    
    if not any((args.init, args.reach, args.machine, args.deadlock,
                args.compatibility, args.agefi, args.sagefi, args.fair_reach,
                args.invariants, args.ctl)):
        args.init = True
        args.reach = True
        args.machine = True
        args.deadlock = True
        args.compatibility = True
        args.agefi = True
        args.sagefi = True
        args.fair_reach = True
        args.invariants = True
        args.ctl = True
    
    with pn.init.init_nusmv():
        MODELFILE = args.model
        
        if args.dynamic:
            if args.verbose:
                verbose("Enabling dynamic BDD reordering")
            pn.nusmv.opt.opt.set_dynamic_reorder(pn.nusmv.opt.opt
                                                 .OptsHandler_get_instance())
        
        if args.verbose:
            verbose("Loading", MODELFILE +
                    (" with ordering in " + args.ordering
                     if args.ordering else ""))
        pn.glob.load(MODELFILE)
        pn.glob.compute_model(variables_ordering=args.ordering)
        fsm = pn.glob.prop_database().master.bddFsm
        
        if args.verbose:
            verbose("The model is encoded with %d binary variables" %
                    fsm.bddEnc.DDmanager.size)
        
        NBROUTES = nbroutes(fsm)
        
        if args.init:
            if args.verbose:
                verbose("Counting and printing initial states")
            init = fsm.init & fsm.state_constraints
            nbinit = fsm.count_states(init)
            print("%d initial state%s" % (nbinit, "s" if nbinit > 1 else ""))
            for state in fsm.pick_all_states(init):
                print("----- Initial state -----")
                for var, val in sorted(state.get_str_values().items()):
                    print(var, "=", val)
        
        if args.reach:
            if args.verbose:
                verbose("Computing the reachable state-space")
            reach = fsm.reachable_states
            print("%d reachable states" % fsm.count_states(reach))
        else:
            reach = None
        
        if args.machine:
            if args.verbose:
                verbose("Checking the fsm consistency")
            pn.nusmv.fsm.bdd.bdd.BddFsm_check_machine(fsm._ptr)
        
        if args.deadlock:
            if args.verbose:
                verbose("Checking the model for deadlocks")
            deadlocks = compute_deadlock_states(fsm)
            if deadlocks.is_false():
                print("The model is deadlock-free")
            else:
                print("The model is not deadlock-free")
                for state in fsm.pick_all_states(deadlocks):
                    print("----- Deadlock state -----")
                    for var, val in sorted(state.get_str_values().items()):
                        print(var, "=", val)
        
        if args.compatibility:
            if args.verbose:
                verbose("Computing compatibility table")
            if reach is None:
                reach = fsm.reachable_states
            setted_routes = extract_routes(fsm, NBROUTES, reach)
            print_compatibility_table(compatibility_table(setted_routes))
        
        if args.agefi:
            if args.verbose:
                verbose("Checking AG EF init")
            snagefi = compute_not_ag_ef_init(fsm, ignore_fairness=
                                                  args.ignore_fairness)
            agefi = snagefi.is_false()
            print("AG EF init is", agefi)
        else:
            agefi = False
        
        if args.sagefi:
            if args.verbose:
                verbose("Computing states for AG EF INIT")
            snagefi = compute_not_ag_ef_init(fsm, ignore_fairness=
                                                  args.ignore_fairness)
            if snagefi.is_false():
                agefi = True
                print("AG EF INIT is True, no violating states")
            else:
                print("AG EF INIT is False")
                for state in fsm.pick_all_states(snagefi):
                    print("----- State violating AG EF INIT -----")
                    for var, val in sorted(state.get_str_values().items()):
                        print(var, "=", val)
        
        if args.fair_reach:
            if args.verbose:
                verbose("Checking that all reachable states are fair")
            reach = fsm.reachable_states
            if len(fsm.fairness_constraints) <= 0:
                fair = pn.dd.BDD.true()
            else:
                fair = fsm.fair_states
            unfair_reach = ((reach - fair) & fsm.state_constraints &
                            fsm.bddEnc.statesMask)
            if unfair_reach.is_false():
                print("All reachable states are fair")
            else:
                print("Some reachable states are not fair")
                for state in fsm.pick_all_states(unfair_reach):
                    print("----- Unfair reachable state -----")
                    for var, val in sorted(state.get_str_values().items()):
                        print(var, "=", val)
        
        if args.invariants or args.ctl:
            if args.verbose:
                verbose("Checking properties")
                if args.ignore_fairness:
                    verbose("Ignoring fairness constraints "
                            "(%d fairness constraints ignored)"
                            % len(fsm.fairness_constraints))
            for prop in pn.glob.prop_database():
                if (prop.type == pn.prop.propTypes['Invariant'] and
                    args.invariants):
                    if args.verbose:
                        verbose("Checking INVARSPEC", prop.expr)
                    print("INVARSPEC", prop.expr,
                          "is", check_invariant(fsm, str(prop.expr)))
                if (prop.type == pn.prop.propTypes['CTL'] and args.ctl):
                    if args.verbose:
                        verbose("Checking CTLSPEC", prop.expr)
                    print("CTLSPEC", prop.expr,
                          "is", check_ctl(fsm, prop.expr, agefi,
                                          ignore_fairness=
                                          args.ignore_fairness))
        
        if args.verbose:
            verbose("End of verification")
            if args.dynamic:
                verbose("%d BDD reorderings done" %
                        fsm.bddEnc.DDmanager.reorderings)